/*
 * Created on 18-Jul-2003
 */
package applications;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.util.StringTokenizer;

import protocols.bestEffortBroadcast.BEBLayer;
import protocols.indulgentUniformRB.IURBLayer;
import protocols.perfectFailureDetector.PerfectFailureDetectorLayer;
import protocols.probabilisticBroadcast.PBLayer;
import protocols.reliableBroadcast.RBLayer;
import protocols.sampleAppl.SampleApplLayer;
import protocols.sampleAppl.SampleApplSession;
import protocols.uniformReliableBroadcast.URBLayer;

import appia.Appia;
import appia.AppiaCursorException;
import appia.AppiaDuplicatedSessionsException;
import appia.AppiaInvalidQoSException;
import appia.Channel;
import appia.ChannelCursor;
import appia.Layer;
import appia.QoS;
import appia.protocols.tcpcomplete.TcpCompleteLayer;
import appia.protocols.common.InetWithPort;
import appia.protocols.udpsimple.UdpSimpleLayer;

import utils.ProcessSet;
import utils.SampleProcess;

/**
 * This class is the MAIN class to run the Reliable Broadcast protocols. 
 * @author nuno
 */
public class SampleAppl {

	/**
	 * Builds the Process set, using the information in the specified file.
	 * @param filename the location of the file
	 * @param selfProc the number of the self process
	 * @return a new ProcessSet
	 */
	private static ProcessSet buildProcessSet(String filename, int selfProc) {
		BufferedReader reader = null;
		try {
			reader =
				new BufferedReader(
					new InputStreamReader(new FileInputStream(filename)));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.exit(0);
		}
		String line = null;
		StringTokenizer st = null;
		boolean hasMoreLines = true;
		int numProcesses=-1;
		try {
			line = reader.readLine();
			numProcesses = Integer.parseInt(line);
		} catch (NumberFormatException e1) {
			System.err.println("First line must be the number of processes.");
			System.exit(1);
		} catch (IOException e1) {
			System.err.println("Error reading the file: "+e1.getMessage());
			System.exit(1);
		}
		ProcessSet set = new ProcessSet(numProcesses);
//		reads lines of type: <process number> <IP address> <port>
		for (int i=0; hasMoreLines && i<numProcesses; i++) {
			try {
				line = reader.readLine();
				st = new StringTokenizer(line);
				if (st.countTokens() != 3) {
					System.err.println("Wrong line in file.");
					continue;
				}
				int procNumber = Integer.parseInt(st.nextToken());
				InetAddress addr = InetAddress.getByName(st.nextToken());
				int portNumber = Integer.parseInt(st.nextToken());
				boolean self = (procNumber == selfProc);
				SampleProcess process =
					new SampleProcess(
						new InetWithPort(addr, portNumber),
						procNumber,
						self);
				set.addProcess(process,procNumber);
			} catch (IOException e) {
				hasMoreLines = false;
			} catch (NumberFormatException e) {
				System.err.println("Wrong line in file.");
				continue;
			}
		} // end of while
		return set;
	}

	/**
	 * Builds an Appia channel with the specified QoS
	 * @param set the ProcessSet
	 * @param qos the specified QoS
	 * @return a new uninitialized channel
	 */
	private static Channel getChannel(ProcessSet set, String qos) {
		if(qos.equals("beb"))
			return getBebChannel(set);
		else if(qos.equals("rb"))
			return getRbChannel(set);
		else if(qos.equals("urb"))
			return getURbChannel(set);
		else if(qos.equals("iurb"))
			return getIURbChannel(set);
		else{
			StringTokenizer st = new StringTokenizer(qos);
			if(st.countTokens() != 3)
				invalidArgs();
			if( ! st.nextToken().equals("pb") )
				invalidArgs();
			int fanout = 0;
			int rounds = 0;
			try {
				fanout = Integer.parseInt(st.nextToken());
				rounds = Integer.parseInt(st.nextToken());
			} catch (NumberFormatException e) {
				e.printStackTrace();
				invalidArgs();
			}
			return getPBChannel(set,fanout,rounds);
		} 
	}


	/**
	 * Builds a new Channel with Probabilistic Broadcast.
	 * @param processes set of proccesses
	 * @param fanout fanout to use in the protocol
	 * @param rounds number of rounds to use in the protocol
	 * @return a new uninitialized Channel
	 */
	private static Channel getPBChannel(ProcessSet processes, int fanout, int rounds) {
		/* Creates a new PBLayer and initializes it */
		PBLayer pbLayer = new PBLayer();
		pbLayer.initValues(fanout,rounds);
		/* Create layers and put them on a array */
		Layer[] qos =
			{new UdpSimpleLayer(), pbLayer, new SampleApplLayer()};

		/* Create a QoS */
		QoS myQoS = null;
		try {
			myQoS = new QoS("Probabilistic Broadcast QoS", qos);
		} catch (AppiaInvalidQoSException ex) {
			System.err.println("Invalid QoS");
			System.err.println(ex.getMessage());
			System.exit(1);
		}
		/* Create a channel. Uses default event scheduler. */
		Channel channel = myQoS.createUnboundChannel("Probabilistic Broadcast Channel");
		/* Application Session requires special arguments: filename and .
		   A session is created and binded to the stack. Remaining ones
		   are created by default
		*/
		SampleApplSession sas = (SampleApplSession) qos[qos.length-1].createSession();
		sas.init(processes);
		ChannelCursor cc=channel.getCursor();
		/* Application is the last session of the array. Positioning
		   in it is simple */
		try {
			cc.top();
			cc.setSession(sas);
		} catch(AppiaCursorException ex) {
			System.err.println("Unexpected exception in main. Type code:" + ex.type);
			System.exit(1);
		}
		return channel;
	}

	/**
	 * Builds a new Appia Channel with Best Effort Broadcast
	 * @param processes set of processes
	 * @return a new uninitialized Channel
	 */
	private static Channel getBebChannel(ProcessSet processes) {
		/* Create layers and put them on a array */
		Layer[] qos =
			{new TcpCompleteLayer(), new BEBLayer(), new SampleApplLayer()};

		/* Create a QoS */
		QoS myQoS = null;
		try {
			myQoS = new QoS("Best Effort Broadcast QoS", qos);
		} catch (AppiaInvalidQoSException ex) {
			System.err.println("Invalid QoS");
			System.err.println(ex.getMessage());
			System.exit(1);
		}
		/* Create a channel. Uses default event scheduler. */
		Channel channel = myQoS.createUnboundChannel("Best effort Broadcast Channel");
		/* Application Session requires special arguments: filename and .
		   A session is created and binded to the stack. Remaining ones
		   are created by default
		*/
		SampleApplSession sas = (SampleApplSession) qos[qos.length-1].createSession();
		sas.init(processes);
		ChannelCursor cc=channel.getCursor();
		/* Application is the last session of the array. Positioning
		   in it is simple */
		try {
			cc.top();
			cc.setSession(sas);
		} catch(AppiaCursorException ex) {
			System.err.println("Unexpected exception in main. Type code:" + ex.type);
			System.exit(1);
		}
		return channel;
	}

	/**
	 * Builds a new Appia Channel with Reliable Broadcast
	 * @param processes set of processes
	 * @return a new uninitialized Channel
	 */
	private static Channel getRbChannel(ProcessSet processes) {
		/* Create layers and put them on a array */
		Layer[] qos =
			{new TcpCompleteLayer(), 
				new BEBLayer(), 
				new PerfectFailureDetectorLayer(),
				new RBLayer(), 
				new SampleApplLayer()};

		/* Create a QoS */
		QoS myQoS = null;
		try {
			myQoS = new QoS("Reliable Broadcast QoS", qos);
		} catch (AppiaInvalidQoSException ex) {
			System.err.println("Invalid QoS");
			System.err.println(ex.getMessage());
			System.exit(1);
		}
		/* Create a channel. Uses default event scheduler. */
		Channel channel = myQoS.createUnboundChannel("Reliable Broadcast Channel");
		/* Application Session requires special arguments: filename and .
		   A session is created and binded to the stack. Remaining ones
		   are created by default
		*/
		SampleApplSession sas = (SampleApplSession) qos[qos.length-1].createSession();
		sas.init(processes);
		ChannelCursor cc=channel.getCursor();
		/* Application is the last session of the array. Positioning
		   in it is simple */
		try {
			cc.top();
			cc.setSession(sas);
		} catch(AppiaCursorException ex) {
			System.err.println("Unexpected exception in main. Type code:" + ex.type);
			System.exit(1);
		}
		return channel;
	}

	/**
	 * Builds a new Appia Channel with Uniform Reliable Broadcast
	 * @param processes set of processes
	 * @return a new uninitialized Channel
	 */
	private static Channel getURbChannel(ProcessSet processes) {
		/* Create layers and put them on a array */
		Layer[] qos =
			{new TcpCompleteLayer(), 
				new BEBLayer(), 
				new PerfectFailureDetectorLayer(),
				new URBLayer(), 
				new SampleApplLayer()};

		/* Create a QoS */
		QoS myQoS = null;
		try {
			myQoS = new QoS("Uniform Reliable Broadcast QoS", qos);
		} catch (AppiaInvalidQoSException ex) {
			System.err.println("Invalid QoS");
			System.err.println(ex.getMessage());
			System.exit(1);
		}
		/* Create a channel. Uses default event scheduler. */
		Channel channel = myQoS.createUnboundChannel("Uniform Reliable Broadcast Channel");
		/* Application Session requires special arguments: filename and .
		   A session is created and binded to the stack. Remaining ones
		   are created by default
		*/
		SampleApplSession sas = (SampleApplSession) qos[qos.length-1].createSession();
		sas.init(processes);
		ChannelCursor cc=channel.getCursor();
		/* Application is the last session of the array. Positioning
		   in it is simple */
		try {
			cc.top();
			cc.setSession(sas);
		} catch(AppiaCursorException ex) {
			System.err.println("Unexpected exception in main. Type code:" + ex.type);
			System.exit(1);
		}
		return channel;
	}

	/**
	 * Builds a new Appia Channel with Indulgent Uniform Reliable Broadcast
	 * @param processes set of processes
	 * @return a new uninitialized Channel
	 */
	private static Channel getIURbChannel(ProcessSet processes) {
		/* Create layers and put them on a array */
		Layer[] qos =
			{new TcpCompleteLayer(), 
				new BEBLayer(), 
				new IURBLayer(), 
				new SampleApplLayer()};

		/* Create a QoS */
		QoS myQoS = null;
		try {
			myQoS = new QoS("Indulgent Uniform Reliable Broadcast QoS", qos);
		} catch (AppiaInvalidQoSException ex) {
			System.err.println("Invalid QoS");
			System.err.println(ex.getMessage());
			System.exit(1);
		}
		/* Create a channel. Uses default event scheduler. */
		Channel channel = myQoS.createUnboundChannel("Indulgent Uniform Reliable Broadcast Channel");
		/* Application Session requires special arguments: filename and .
		   A session is created and binded to the stack. Remaining ones
		   are created by default
		*/
		SampleApplSession sas = (SampleApplSession) qos[qos.length-1].createSession();
		sas.init(processes);
		ChannelCursor cc=channel.getCursor();
		/* Application is the last session of the array. Positioning
		   in it is simple */
		try {
			cc.top();
			cc.setSession(sas);
		} catch(AppiaCursorException ex) {
			System.err.println("Unexpected exception in main. Type code:" + ex.type);
			System.exit(1);
		}
		return channel;
	}

	private static final int NUM_ARGS = 8;

	/**
	 * Main!
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length < (NUM_ARGS-2)) {
			invalidArgs();
		}

		/* Parse arguments */
		int arg = 0, self = -1;
		String filename = null, qos = null;
		try {
			while(arg < args.length){
				if(args[arg].equals("-f")){
					arg++;
					filename = args[arg]; 
					System.out.println("Reading from file: "+filename);		
				}
				else if(args[arg].equals("-n")){
					arg++;
					try {
						self = Integer.parseInt(args[arg]);
						System.out.println("Process number: "+self);
					} catch (NumberFormatException e) {
						invalidArgs();
					}				
				}
				else if(args[arg].equals("-qos")){
					arg++;
					qos = args[arg];
					if(qos.equals("pb")){
						qos = qos + " " + args[++arg] + " " + args[++arg];
					}
					System.out.println("Starting with QoS: "+qos);
				}
				else
					invalidArgs();
				arg++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			e.printStackTrace();
			invalidArgs();
		}
		
		/* gets a new uninitialized Channel with the specified QoS and the Appl session 
		 * created. Remaining sessions are created by default. Just tell the channel to start. */
		Channel channel = getChannel(buildProcessSet(filename, self), qos);
		try {
			channel.start();
		} catch(AppiaDuplicatedSessionsException ex) {
			System.err.println("Sessions binding strangely resulted in "+
					   "one single sessions occurring more than "+
					   "once in a channel");
			System.exit(1);
		}
	
		/* All set. Appia main class will handle the rest */
		System.out.println("Starting Appia...");
		Appia.run();
	}

	/**
	 * Prints a error message and exit.
	 */
	private static void invalidArgs() {
		System.out.println("Invalid args.\nUsage SampleAppl -f filemane -n proc_number -qos QoS_type."+
		"\nQoS can be one of the following:"+
		"\nbeb - Best Effort Broadcast"+
		"\nrb - Reliable Broadcast"+
		"\nurb - Uniform Reliable Broadcast"+
		"\niurb - Indulgent Uniform Reliable Broadcast"+
		"\npb <f> <r> - Probabilistic Broadcast with a fanout f and a number of rounds r");
		System.exit(1);
	}
}
