/*
 * IURBSession.java
 * Created on 19-Sep-2003, 11:37:01
 */
package protocols.indulgentUniformRB;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

import protocols.utils.MessageEntry;
import protocols.utils.MessageID;

import utils.Debug;
import utils.ProcessSet;
import utils.SampleProcess;
import events.ProcessInitEvent;
import appia.AppiaEventException;
import appia.Direction;
import appia.Event;
import appia.Layer;
import appia.Session;
import appia.events.SendableEvent;
import appia.events.channel.ChannelInit;
import appia.message.ExtendedMessage;
import appia.protocols.common.InetWithPort;

/**
 * Class IURBSession
 * @author nuno
 */
public class IURBSession extends Session {

	// List of processes.
	private ProcessSet processes;
	// This sequence number represents the delivered set.
	private int seqNumber;
	// List of MessageID objects
	private LinkedList forward, delivered;
	// hashtable mapping: < MessageID --> MessageEntry >
	private Hashtable ack;
	

	/**
	 * @param layer
	 */
	public IURBSession(Layer layer) {
		super(layer);
	}

	public void handle(Event event) {
		// Init events. Channel Init is from Appia and ProcessInitEvent is to know the elements of the group
		if (event instanceof ChannelInit)
			handleChannelInit((ChannelInit) event);
		else if (event instanceof ProcessInitEvent)
			handleProcessInitEvent((ProcessInitEvent) event);
		else if (event instanceof SendableEvent) {
			if (event.getDir() == Direction.DOWN)
				// UPON event from the above protocol (or application)
				iurbBroadcast((SendableEvent) event);
			else
				// UPON event from the bottom protocol (or perfect point2point links)
				bebDeliver((SendableEvent) event);
		}
		/*
		 * Unexpected event arrived. Forwarding it.
		 */
		else
			try {
				event.go();
			} catch (AppiaEventException e) {
				e.printStackTrace();
			}		

		// Every time something happens, the protocol verify if more messages can be delivered.
		urbTryDeliver();
	}

	/**
	 * verifies if there are messages to be delivered.
	 */
	private void urbTryDeliver() {
		Iterator it = getMessages();
		MessageEntry entry=null;
		// for each message...
		while( it.hasNext() ){
			entry = (MessageEntry) it.next();
			if(canDeliver(entry)){
				// add the message to the delivered set and
				delivered.add(entry.messageID);
				// delivers the message.
				iurbDeliver(entry.event, entry.messageID.process);
			}
		}
	}

	/**
	 * Verifies if the message can be delivered.
	 * @param entry
	 * @return
	 */
	private boolean canDeliver(MessageEntry entry) {
		int N = processes.getSize(), numAcks = 0;
		for(int i=0; i<N; i++) 
			if(entry.acks[i]) 
				numAcks++;
		return (numAcks > (N/2)) && ( ! delivered.contains(entry.messageID) );
	}

	/**
	 * @param init
	 */
	private void handleChannelInit(ChannelInit init) {
		try {
			init.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
		forward = new LinkedList();
		delivered = new LinkedList();
		ack = new Hashtable();
	}

	/**
	 * @param event
	 */
	private void handleProcessInitEvent(ProcessInitEvent event) {
		processes = event.getProcessSet();
		try {
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Called when the above protocol sends a message.
	 * @param event
	 */
	private void iurbBroadcast(SendableEvent event) {
		/* header is composed by a sequence number
		 * and the number of the process that is sending the message
		 */
		SampleProcess self = processes.getSelfProcess();
		MessageID msgID = new MessageID(self.getProcessNumber(),seqNumber);
		Debug.print("URB: broadcasting message from "+ msgID.process +"with seqNumber = "+msgID.seqNumber);
		seqNumber++; 
		// adds the message to the forward set.
		forward.add(msgID);
		// push the message header
		push(event,msgID);
		// broadcasts the message.
		bebBroadcast(event);
	}

	/**
	 * Called when the lower protocol delivers a message.
	 * @param event
	 */
	private void bebDeliver(SendableEvent event) {
		Debug.print("URB: Received message from beb.");
		SendableEvent clone = null;
		try {
			clone = (SendableEvent) event.cloneEvent();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return;
		}
		// pops the message header
		MessageID msgID = pop(clone);
		// register the information about the ack
		addAck(clone,msgID);
		if( ! forward.contains(msgID) ){
			Debug.print("URB: Message is not on the forward set.");
			// forwards the message
			forward.add(msgID);
			bebBroadcast(event);
		}
	}

	/**
	 * Called by this protocol to send a message to the lower protocol.
	 * @param event
	 */
	private void bebBroadcast(SendableEvent event) { 
		Debug.print("URB: sending message to beb.");
		try {
			if(event.getDir() != Direction.DOWN)
				event.setDir(Direction.invert(event.getDir()));
			event.setSource(this);
			event.init();
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Delivers the message to above protocol or application
	 * @param event
	 * @param self
	 */
	private void iurbDeliver(SendableEvent event, int sender) {
		Debug.print("URB: delivering message to above protocol.");
		try {
			if(event.getDir() != Direction.UP)
				event.setDir(Direction.invert(event.getDir()));
			event.setSource(this);
			event.source = processes.getProcess(sender).getInetWithPort();
			event.init();
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}
	
	/*
	 * ###################################
	 * Methods to help the protocol
	 * ###################################
	 */
	 
	private void push(SendableEvent event, MessageID msgID){
		ExtendedMessage message = (ExtendedMessage) event.getMessage();
		message.pushObject(msgID);
	}
	
	private MessageID pop(SendableEvent event){
		ExtendedMessage message = (ExtendedMessage) event.getMessage();
		return (MessageID) message.popObject();
	}
	
	private void addAck(SendableEvent event, MessageID msgID){
		Debug.print("URB: adding ack.");
		int pi = processes.getProcess((InetWithPort)event.source).getProcessNumber();
		MessageEntry entry = (MessageEntry) ack.get(msgID);
		if(entry == null){
			Debug.print("URB: first time that the message is seen.");
			entry = new MessageEntry(event, msgID, processes.getSize());
			ack.put(msgID,entry);
		}
		entry.acks[pi] = true;
	}

	private MessageEntry getAck(MessageID msgID){
		return (MessageEntry) ack.get(msgID);
	}
	
	private Iterator getMessages(){
		return ack.values().iterator();
	}
}


