/*
 * PBSession.java
 * Created on 22-Sep-2003, 17:09:41
 */
package protocols.probabilisticBroadcast;

import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Random;

import protocols.utils.MessageID;

import utils.Debug;
import utils.ProcessSet;
import utils.SampleProcess;

import events.ProcessInitEvent;
import appia.AppiaEventException;
import appia.Direction;
import appia.Event;
import appia.Layer;
import appia.Session;
import appia.events.SendableEvent;
import appia.events.channel.ChannelInit;
import appia.message.ExtendedMessage;
import appia.protocols.common.InetWithPort;

/**
 * Class PBSession
 * @author nuno
 */
public class PBSession extends Session {

	// list of delivered messages
	private LinkedList delivered;
	// set of processes
	private ProcessSet processes;
	private int fanout, maxRounds, seqNumber; 

	/**
	 * @param layer
	 */
	public PBSession(Layer layer) {
		super(layer);
		PBLayer pbLayer = (PBLayer) layer;
		fanout = pbLayer.getFanout();
		maxRounds = pbLayer.getMaxRounds();
		seqNumber = 0;
	}

	public void handle(Event event){
		// Init events. Channel Init is from Appia and ProcessInitEvent is to know the elements of the group
		if(event instanceof ChannelInit)
			handleChannelInit((ChannelInit)event);
		else if(event instanceof ProcessInitEvent)
			handleProcessInitEvent((ProcessInitEvent) event);
		else if(event instanceof SendableEvent){
			if(event.getDir()==Direction.DOWN)
				// UPON event from the above protocol (or application)
				pbBroadcast((SendableEvent) event);
			else
				// UPON event from the bottom protocol (or perfect point2point links)
				up2pDeliver((SendableEvent) event);
		}
	}

	/**
	 * @param init
	 */
	private void handleChannelInit(ChannelInit init) {
		try {
			init.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
		delivered = new LinkedList();
	}

	/**
	 * @param event
	 */
	private void handleProcessInitEvent(ProcessInitEvent event) {
		processes = event.getProcessSet();
		fanout = Math.min (fanout, processes.getSize());
		try {
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}		
	}

	/**
	 * @param event
	 */
	private void pbBroadcast(SendableEvent event) {
		MessageID msgID = new MessageID(processes.getSelfRank(),seqNumber);
		seqNumber++;
		// gossips a new message.
		gossip(event, msgID, maxRounds-1);
	}

	/**
	 * @param event
	 */
	private void up2pDeliver(SendableEvent event) {
		SampleProcess pi = processes.getProcess((InetWithPort)event.source);
		// pops the round of the message
		int round = ((ExtendedMessage) event.getMessage()).popInt();
		// and the remaining header
		MessageID msgID = (MessageID) ((ExtendedMessage) event.getMessage()).popObject();
		// if the message was not delivered yet
		if( ! delivered.contains(msgID) ){
			delivered.add(msgID);
			
			SendableEvent clone = null;
			try {
				clone = (SendableEvent) event.cloneEvent();
			} catch (CloneNotSupportedException e) {
				e.printStackTrace();
			}
			// delivers the message
			pbDeliver(clone,msgID);
		}
		
		// gossips the message again, if there are more rounds.
		if(round > 0)
			gossip(event,msgID,round-1);
	}

	private void gossip(SendableEvent event, MessageID msgID, int round){
			Debug.print("PB: gossip of message"+msgID+" in round "+round);
			int[] targets = pickTargets();
			for(int i=0; i<fanout; i++){
				SendableEvent clone = null;
				try {
					clone = (SendableEvent) event.cloneEvent();
				} catch (CloneNotSupportedException e) {
					e.printStackTrace();
				}
				((ExtendedMessage) clone.getMessage()).pushObject(msgID);
				((ExtendedMessage) clone.getMessage()).pushInt(round);
				up2pSend(clone,targets[i]);
			}
	}

	/**
	 * This method returns n diferent process ranks,
	 * where n is the fanout.
	 */
	private int[] pickTargets() {
		Random random = new Random(System.currentTimeMillis());
		LinkedList targets = new LinkedList();
		Integer candidate = null;
		while(targets.size() < fanout){
			candidate = new Integer(random.nextInt(processes.getSize()));	
			if( ( ! targets.contains(candidate) ) && (candidate.intValue() != processes.getSelfRank()) )
				targets.add(candidate);
		}
		int[] targetArray = new int[fanout];
		ListIterator it = targets.listIterator();
		for(int i=0; (i<targetArray.length) && it.hasNext(); i++)
			targetArray[i] =  ((Integer)it.next()).intValue();
		return targetArray;
	}

	/**
	 * Called by this protocol to send a message to the lower protocol.
	 * @param event
	 */
	private void up2pSend(SendableEvent event, int dest) { 
		Debug.print("PB: sending message to up2p.");
		try {
			event.setDir(Direction.DOWN);
			event.setSource(this);
			event.dest = processes.getProcess(dest).getInetWithPort();
			event.init();
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Delivers the message to above protocol or application
	 * @param event
	 * @param self
	 */
	private void pbDeliver(SendableEvent event, MessageID msgID) {
		Debug.print("PB: delivering message to above protocol.");
		try {
			event.setDir(Direction.UP);
			event.setSource(this);
			event.source = processes.getProcess(msgID.process).getInetWithPort();
			event.init();
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}

}
