/*
 * @author Nuno Carvalho
 * Created on 18-Jul-2003
 *
 */
package protocols.reliableBroadcast;

import java.util.LinkedList;
import java.util.ListIterator;

import protocols.utils.MessageID;

import utils.Debug;
import utils.ProcessSet;
import utils.SampleProcess;
import events.Crash;
import events.ProcessInitEvent;
import appia.AppiaEventException;
import appia.Direction;
import appia.Event;
import appia.Layer;
import appia.Session;
import appia.events.SendableEvent;
import appia.events.channel.ChannelInit;
import appia.message.ExtendedMessage;
import appia.protocols.common.InetWithPort;

/**
 * @author nuno
 *
 */
public class RBSession extends Session {

	private ProcessSet processes;
	private int seqNumber;
	// array of lists
	private LinkedList[] from;
	// List of MessageID objects
	private LinkedList delivered;

	/**
	 * @param layer
	 */
	public RBSession(Layer layer) {
		super(layer);
		seqNumber = 0;
	}

/**
 * Main event handler
 */
	public void handle(Event event){
		// Init events. Channel Init is from Appia and ProcessInitEvent is to know the elements of the group
		if(event instanceof ChannelInit)
			handleChannelInit((ChannelInit)event);
		else if(event instanceof ProcessInitEvent)
			handleProcessInitEvent((ProcessInitEvent) event);
		else if(event instanceof SendableEvent){
			if(event.getDir()==Direction.DOWN)
				// UPON event from the above protocol (or application)
				rbBroadcast((SendableEvent) event);
			else
				// UPON event from the bottom protocol (or perfect point2point links)
				bebDeliver((SendableEvent) event);
		}
		else if(event instanceof Crash)
			handleCrash((Crash)event);
	}

	/**
	 * @param init
	 */
	private void handleChannelInit(ChannelInit init) {
		try {
			init.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
		delivered = new LinkedList();
	}

	/**
	 * @param event
	 */
	private void handleProcessInitEvent(ProcessInitEvent event) {
		processes = event.getProcessSet();
		try {
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
		from = new LinkedList[processes.getSize()];
		for (int i=0; i<from.length; i++)
			from[i] = new LinkedList();
	}

	/**
	 * Called when the above protocol sends a message.
	 * @param event
	 */
	private void rbBroadcast(SendableEvent event) {
		// first we take care of the header of the message
		SampleProcess self = processes.getSelfProcess();
		MessageID msgID = new MessageID(self.getProcessNumber(),seqNumber);
		seqNumber++; 
		Debug.print("RB: broadcasting message.");
		((ExtendedMessage)event.getMessage()).pushObject(msgID);
		// broadcast the message
		bebBroadcast(event);
	}

	/**
	 * Called when the lower protocol delivers a message.
	 * @param event
	 */
	private void bebDeliver(SendableEvent event) {
		Debug.print("RB: Received message from beb.");
		MessageID msgID = (MessageID) ((ExtendedMessage)event.getMessage()).peekObject();
		if( ! delivered.contains(msgID) ){
			Debug.print("RB: message is new.");
			delivered.add(msgID);
			// removes the header from the message (sender and seqNumber) and delivers it
			SendableEvent cloned = null;
			try {
				cloned = (SendableEvent) event.cloneEvent();
			} catch (CloneNotSupportedException e) {
				e.printStackTrace();
				return;
			}
			((ExtendedMessage)event.getMessage()).popObject();
			try {
			    event.go();
			} catch (AppiaEventException e) {
			    e.printStackTrace();
			}
			// adds message to the "from" array
			SampleProcess pi = processes.getProcess((InetWithPort) event.source);
			int piNumber = pi.getProcessNumber(); 
			from[piNumber].add(cloned);
			/*
			 * resends the message if the source is no longer correct
			 */
			if( ! pi.isCorrect() ){
			        SendableEvent retransmission = null;

				try {
					retransmission = (SendableEvent) cloned.cloneEvent();
				} catch (CloneNotSupportedException e1) {
					e1.printStackTrace();
				}
				bebBroadcast(retransmission);
			}
		}
	}

	/**
	 * Called by this protocol to send a message to the lower protocol.
	 * @param event
	 */
	private void bebBroadcast(SendableEvent event) { 
		Debug.print("RB: sending message to beb.");
		try {
			event.setDir(Direction.DOWN);
			event.setSource(this);
			event.init();
			event.go();
		} catch (AppiaEventException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Called when some process crashed.
	 * @param crash
	 */
	private void handleCrash(Crash crash) {
		int pi = crash.getCrashedProcess();
		System.out.println("Process "+pi+" failed.");
		
		// changes the state of the process to "failed"
		processes.getProcess(pi).setCorrect(false);
		
		// resends the messages of the failed process
		SendableEvent event = null;
		ListIterator it = from[pi].listIterator();
		while(it.hasNext()){
			event = (SendableEvent) it.next();
			bebBroadcast(event);
		}
	}
}
